#ifndef lint
static char *RCSid = "@(#)$Id: xmexit.c,v 2.0 90/04/09 10:03:42 tot Exp $";
#endif /* not lint */

/* Program to shut down X server. Based on xmexit by Teemu Torma.

   Copyright 1990 Front End Oy

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation, and that the name of Front End not be used in advertising
   or publicity pertaining to distribution of the software without specific,
   written prior permission.  Front End makes no representations about the
   suitability of this software for any purpose.  It is provided "as is"
   without express or implied warranty.

   FRONT END DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
   ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
   FRONT END.  BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
   ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
   IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
   OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:  Teemu Torma, Front End Oy
            tot@frend.fi

   Ported to Athena Widgets by: Paul Raines
				raines@slac.stanford.edu

*/

#include <X11/X.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Shell.h>
#include <stdio.h>
#include <signal.h>
#include "xtexit.h"

/* Application resources. */

static XtResource Resources[] = {
  { Nconfirm, CConfirm, XtRBoolean, sizeof (Boolean),
      XtOffset (XTexit, confirm), XtRImmediate, (caddr_t) True },
  { Ndebug, CDebug, XtRBoolean, sizeof (Boolean),
      XtOffset (XTexit, debug), XtRImmediate, (caddr_t) False },
  { Ndelay, CDelay, XtRInt, sizeof (int),
      XtOffset (XTexit, delay), XtRImmediate, (caddr_t) 2 },
  { Nlevels, CLevels, XtRInt, sizeof (int),
      XtOffset (XTexit, levels), XtRImmediate, (caddr_t) 0 },
  { Ntext, CText, XtRString, sizeof (char *),
      XtOffset (XTexit, text), XtRString, (XtPointer) "Exit X-Window?" },
};

/* Options for us. */

static XrmOptionDescRec Options[] = {
  { "-quiet", "*confirm", XrmoptionNoArg, "False" },
  { "-confirm", "*confirm", XrmoptionNoArg, "True" },
  { "-text", "*text", XrmoptionSepArg },
  { "-levels", "*levels", XrmoptionSepArg },
  { "-delay", "*delay", XrmoptionSepArg },
  { "-debug", "*debug", XrmoptionNoArg, "True" },
};

/* Fallback resources */

static String fallback_resources[] = {
    "*foreground:   black",
    "*background:   lightblue",
    "*title:	    Confirm",
    "*Geometry:     180x70",
    "*Command*font: -adobe-helvetica-bold-r-normal--11-*-*-*-*-*-*-*",
    "*Label*font:    -b&h-lucidatypewriter-medium-r-normal-*-12-*-*-*-*-*-*-*",
    "*Command*shapeStyle:       oval",
    NULL
};

/* Translations */

static void KeyConfirm(), KeyCancel();

static char defaultTranslations[] =
  "<Key> Return:   confirm()\n\
   <Key> Escape:   cancel()";

static XtActionsRec window_actions[] = {
  {"confirm", KeyConfirm},
  {"cancel", KeyCancel},
};


/* Atoms we need to use. */

static Atom WmState;
static Atom WmProtocols;
static Atom WmDeleteWindow;

/* If set, we are debugging.
   Just show actions we should make. */

static Boolean Debug = False;

static XTexit p_xtexit;


/* Show usage of this program. */

static void
Usage ()
{
  fputs ("\
usage:\n\
        xtexit [-options]\n\
\n\
where options include X toolkit options and\n\
    -quiet                  Do not show confirm dialog\n\
    -confirm                Do show confirm dialog (default)\n\
    -text <string>          Text to show in dialog\n\
    -levels <num>           Look applications only NUM levels downs\n\
    -delay <num>            Wait NUM seconds before force kill\n\
", stderr);
  exit (1);
}


/* Delete WINDOW of DISPLAY by using WM_DELETE_WINDOW protocol. */

static void
DeleteWindow (display, window)
     Display *display;
     Window window;
{
  XClientMessageEvent event;

  event.display = display;
  event.window = window;
  event.type = ClientMessage;
  event.send_event = True;
  event.message_type = WmProtocols;
  event.format = 32;
  event.data.l[0] = WmDeleteWindow;

  XSendEvent (display, window, False, 0, (XEvent *) &event);
}


/* See if WINDOW has property PROP.
   Return True if it has, False if doesn't */

static Boolean
PropertyExist (display, window, prop)
     Display *display;
     Window window;
     Atom prop;
{
  int num_props, count;
  Atom *props;

  props = XListProperties (display, window, &num_props);
  if (props)
    {
      for (count = 0; count < num_props; count ++)
	if (props[count] == prop)
	  {
	    XFree ((char *) props);
	    return True;
	  }
      XFree ((char *) props);
    }
  return False;
}


/* Kill application for WINDOW.
   If FORCE, always do XKillClient, otherwise try to use WM_DELETE_WINDOW
   protocol if window has one. */

static void
KillApplication (xtexit, window, force)
     XTexit xtexit;
     Window window;
     Boolean force;
{
  Display *display = xtexit->display;
  Atom *protocols;
  int num_protocols;

  /* If this is our window, skip. */

  if (window == xtexit->top
      || (xtexit->dialog && window == xtexit->dialog))
    {
      if (Debug)
	printf ("Skipping window %#8x as own window\n", window);
      return;
    }

  /* If window has WM_PROCOTOLS property, see if it contains WM_DELETE_WINDOW
     property. */

  if (! force && PropertyExist (display, window, WmProtocols)
      && XGetWMProtocols (display, window, &protocols, &num_protocols))
    {
      int count;
      Boolean found = False;

      for (count = 0; count < num_protocols; count ++)
	if (protocols[count] == WmDeleteWindow)
	  {
	    found = True;
	    break;
	  }

      if (found)
	{
	  if (Debug)
	    printf ("Kill window %#8x using WM_DELETE_WINDOW\n", window);
	  else
	    DeleteWindow (display, window);
	}
      else
	{
	  if (Debug)
	    printf ("Kill window %#8x using XKillClient\n", window);
	  else
	    XKillClient (display, window);
	}

      XFree ((char *) protocols);
    }
  else if (Debug)
    printf ("Kill window %#8x using XKillClient\n", window);
  else
    XKillClient (display, window);
}      


/* Add WINDOW to xtexit's window list. */

static void
AddWindowToList (xtexit, window)
     XTexit xtexit;
     Window window;
{
  if (! xtexit->window_list)
    {
      xtexit->window_list = (Window *) XtMalloc (100 * sizeof (Window));
      xtexit->window_list_size = 100;
      xtexit->last_window = 0;
    }
  else if (xtexit->last_window >= xtexit->window_list_size)
    {
      xtexit->window_list_size += 100;
      xtexit->window_list = (Window *)
	XtRealloc ((char *) xtexit->window_list,
		   xtexit->window_list_size * sizeof (Window));
    }

  xtexit->window_list[xtexit->last_window ++] = window;
}


/* Our own error handler.
   If error is BadWindow in certain of our requests, ignore it. */

static int
ErrorHandler (display, error)
     Display *display;
     XErrorEvent *error;
{
  switch (error->request_code)
    {
    case X_QueryTree:
    case X_ListProperties:
    case X_SendEvent:
    case X_KillClient:
    case X_GetProperty:

      /* Ignore BadWindow and BadValue errors since we will get them.
	 It's because killing windows usually cause destroying other windows
	 also since normally clients have more than just one window. */

      if (error->error_code == BadValue || error->error_code == BadWindow)
	return 0;
      break;
    }

  return _XDefaultError (display, error);
}


/* Read windows to xtexit's window list below ROOT window.
   LEVEL is current level we are at.
   If ALL, read all windows, regardless of LEVEL and whether WM_STATE property
   exists (except our windows, of cource). */

static void
ReadWindows (xtexit, root, level, all)
     XTexit xtexit;
     Window root;
     int level;
     Boolean all;
{
  Display *display = xtexit->display;
  Window dummy, *kids;
  unsigned int num_kids;
  int count;

  /* If we are below maximum level, nothing to do. */

  if (! all && xtexit->levels && level > xtexit->levels)
    return;

  if (! XQueryTree (display, root, &dummy, &dummy, &kids, &num_kids))
    return;

  /* Add kids with WM_STATE property to window list. They are potential X
     clients.
     If not possible X client, recurse tree below it. */

  for (count = 0; count < num_kids; count ++)
    if (kids[count] != xtexit->top
	&& (! xtexit->dialog || kids[count] != xtexit->dialog))
      {
	if (all)
	  {
	    AddWindowToList (xtexit, kids[count]);
	    ReadWindows (xtexit, kids[count], level + 1, all);
	  }
	else if (PropertyExist (display, kids[count], WmState))
	  AddWindowToList (xtexit, kids[count]);
	else
	  ReadWindows (xtexit, kids[count], level + 1, all);
      }
  
  XFree ((char *) kids);
}


/* Kill all the clients from the X server. */
/* kill -15 */

static void
KillClients15 (xtexit)
     XTexit xtexit;
{
  Display *display = xtexit->display;
  int screen, count;

  /* Intern atoms we need. */

  WmState = XInternAtom (display, "WM_STATE", False);
  WmProtocols = XInternAtom (display, "WM_PROTOCOLS", False);
  WmDeleteWindow = XInternAtom (display, "WM_DELETE_WINDOW", False);

  /* Initialize window list variables in the xtexit structure. */

  xtexit->window_list = (Window *) 0;
  xtexit->window_list_size = 0;
  xtexit->last_window = 0;

  /* Ignore hangup signals.
     If we are invoked under xterm or emacs, they will kill us when we kill
     them. I wonder whether we should ignore also SIGTERM's.. */

  signal (SIGHUP, SIG_IGN);

  /* Set our own error handler that ignores BadWindow errors. */

  XSetErrorHandler (ErrorHandler);

  /* Read in top level windows in all screens. */

  for (screen = 0; screen < ScreenCount (display); screen ++)
    {
      ReadWindows (xtexit, RootWindow (display, screen), 1, False);
    }

  /* Kill the windows in the window list. */

  for (count = 0; count < xtexit->last_window; count ++)
    KillApplication (xtexit, xtexit->window_list[count], False);

}


/* Kill all the clients from the X server. */
/* kill -9 */

static void
KillClients9 (xtexit)
     XTexit xtexit;
{
  Display *display = xtexit->display;
  int screen, count;


  for (screen = 0; screen < ScreenCount (display); screen ++)

  /* Intern atoms we need. */

  WmState = XInternAtom (display, "WM_STATE", False);
  WmProtocols = XInternAtom (display, "WM_PROTOCOLS", False);
  WmDeleteWindow = XInternAtom (display, "WM_DELETE_WINDOW", False);

  /* Initialize window list variables in the xtexit structure. */

  xtexit->window_list = (Window *) 0;
  xtexit->window_list_size = 0;
  xtexit->last_window = 0;

  /* Ignore hangup signals.
     If we are invoked under xterm or emacs, they will kill us when we kill
     them. I wonder whether we should ignore also SIGTERM's.. */

  signal (SIGHUP, SIG_IGN);

  /* Set our own error handler that ignores BadWindow errors. */

  XSetErrorHandler (ErrorHandler);

  /* Read in top level windows in all screens. */

  for (screen = 0; screen < ScreenCount (display); screen ++)
    {
      ReadWindows (xtexit, RootWindow (display, screen), 1, False);
    }

  /* Remove our windows from the display.
     Just for the visual effect. */

  XUnmapWindow (display, xtexit->top);
  if (xtexit->dialog)
    XUnmapWindow (display, xtexit->dialog);

  XSync (display, False);

  /* If we've got delay, wait that many seconds before doing force kill.
     This time, we read all the windows and do XKillClient(3X11) to them. */

  /* Remove all windows from the window list. */

  xtexit->last_window = 0;

  /* Read all windows into the memory and kill them. */

  for (screen = 0; screen < ScreenCount (display); screen ++)
  {
    ReadWindows (xtexit, RootWindow (display, screen), 1, True);
  }
  printf("Anzahl clients: %d\n", xtexit->last_window);
  for (count = 0; count < xtexit->last_window; count ++)
  KillApplication (xtexit, xtexit->window_list[count], True);
  XSync (display, False);
}
  

/* Confirm callback which never returns.
   This is called when use confirms  killing all applications. */

static void
ConfirmCallback (w, closure, data)
     Widget w;
     caddr_t closure, data;
{
  XTexit xtexit = (XTexit) closure;

  KillClients9 (xtexit);
  exit (0);
}

/* ARGSUSED */
static void
KeyConfirm (w, event, params, num_params)
Widget w;
XExposeEvent *event;
String *params;
Cardinal *num_params;
{
  KillClients9 (p_xtexit);
  exit (0);
}


/* Cancel callback.
   Just exit the application. */

static void
CancelCallback (w, closure, data)
     Widget w;
     caddr_t closure, data;
{
  XTexit xtexit = (XTexit) closure;

  exit (1);
}

/* ARGSUSED */
static void
KeyCancel (w, event, params, num_params)
Widget w;
XExposeEvent *event;
String *params;
Cardinal *num_params;
{
  exit (0);
}


/* Popup confirm dialog.
   PARENT is parent widget for us, XMEXIT is application data structure. */

static void
PopupConfirmDialog (parent, xtexit)
     Widget parent;
     XTexit xtexit;
{
  Widget shell, box;
  Arg args[20];
  Cardinal num;
  Widget children[20];
  Dimension width, height;
  Position x, y;
  Display *display = XtDisplay (parent);
  XtTranslations mytranslations;
  int screen = XScreenNumberOfScreen (XtScreen (parent));


  num = 0;
  XtSetArg (args[num], XtNallowShellResize, True); num ++;
  shell = XtCreatePopupShell ("confirmDialogShell", transientShellWidgetClass,
			      parent, args, num);

  num = 0;
  XtSetArg (args[num], XtNlabel, xtexit->text); num ++;
  XtSetArg (args[num], XtNresizable, True); num ++;
  box = XtCreateWidget ("confirmDialog", dialogWidgetClass, shell,
			args, num);

  /* Add callbacks for different actions.
     We shouldn't have any other possibilities than OK or Cancel. */

  XawDialogAddButton (box, "Yes", ConfirmCallback, (caddr_t) xtexit);
  XawDialogAddButton (box, "No", CancelCallback, (caddr_t) xtexit);

  /* Destroy callback if user closes dialog.
     Just call confirm callback as we would have cancelled the operation. */

  XtAddCallback (box, XtNdestroyCallback, CancelCallback, (caddr_t) xtexit);

  /* Add translations */

  p_xtexit = xtexit;
  mytranslations = XtParseTranslationTable(defaultTranslations);
  XtOverrideTranslations(box, mytranslations);

  /* Realize the shell and center it to the window. */

  XtManageChild (box);

  XtRealizeWidget (shell);

  num = 0;
  XtSetArg(args[num], XtNheight, &height); num ++;
  XtSetArg(args[num], XtNwidth, &width); num ++;
  XtGetValues (shell, args, num);

  x = DisplayWidth (display, screen) / 2 - (int) width / 2;
  y = DisplayHeight (display, screen) / 2 - (int) height / 2;

  num = 0;
  XtSetArg (args[num], XtNx, x); num ++;
  XtSetArg (args[num], XtNy, y); num ++;
  XtSetValues (shell, args, num);

  XtPopup(shell, XtGrabExclusive);
  
  /* Save dialog shell's window to xtexit structure so that killing will know
     to pass it. */

  xtexit->dialog = XtWindow (shell);
}


int
main (argc, argv)
     int argc;
     char **argv;
{
  Widget top;
  XTexitRec xtexit;
  XtAppContext appcontext;
  int num_clients;

  /* Initialize X toolkit and make top level shell.
     Move this widget to somewhere where it's not visible. */

  top = XtAppInitialize (&appcontext, "XTexit", Options, XtNumber (Options),
		      &argc, argv, fallback_resources, NULL, 0);
  if (argc > 1)
    Usage ();

  XtSetMappedWhenManaged (top, False);
  XtRealizeWidget (top);

  /* Get application resources for us.
     We'll need first to register Xm converters first. */

  XtGetApplicationResources (top, (caddr_t) &xtexit, Resources,
			     XtNumber (Resources), (ArgList) 0, 0);
  Debug = xtexit.debug;

  /* Set our top level windows. */

  xtexit.display = XtDisplay (top);
  xtexit.top = XtWindow (top);
  xtexit.dialog = 0;
  
  /* 1. send a "signal 15" to all clients */
  KillClients15 (&xtexit);

  num_clients = 3;
  if (num_clients < 3)
    KillClients9(&xtexit);

  /* 3. Dialog-Popup and send a "signal 9" to all clients */
  XtAppAddActions(appcontext, window_actions, XtNumber(window_actions));
  PopupConfirmDialog (top, &xtexit);
  XtAppMainLoop (appcontext);

  exit (0);
}
